/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <signal.h>
#include <bsd/string.h>
#include <rte_byteorder.h>

#include <doca_argp.h>
#include <doca_log.h>

#include "firewall_core.h"

DOCA_LOG_REGISTER(FIREWALL);

static bool force_quit;

static void
signal_handler(int signum)
{
	if (signum == SIGINT || signum == SIGTERM) {
		DOCA_LOG_INFO("Signal %d received, preparing to exit...", signum);
		force_quit = true;
	}
}

int
main(int argc, char **argv)
{
	doca_error_t ret;
	const char *grpc_address;
	struct firewall_cfg firewall_cfg = {0};
	struct rule_match *drop_rules;
	int n_rules;

	force_quit = false;

	ret = doca_argp_init("firewall", &firewall_cfg);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to init ARGP resources: %s", doca_get_error_string(ret));
	doca_argp_set_grpc_program();
	register_firewall_params();
	ret = doca_argp_start(argc, argv);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to parse application input: %s", doca_get_error_string(ret));

	if (doca_argp_get_grpc_addr(&grpc_address) != DOCA_SUCCESS)
		APP_EXIT("Failed to get grpc address");

	/* Start the server */
	firewall_ports_init(grpc_address);
	if (firewall_cfg.mode == FIREWALL_MODE_STATIC) {
		drop_rules = init_drop_rules(firewall_cfg.json_path, &n_rules);
		firewall_pipes_init(drop_rules, n_rules);
		signal(SIGINT, signal_handler);
		signal(SIGTERM, signal_handler);
		DOCA_LOG_INFO("Waiting for traffic, press ctrl+c for termination");
		while (!force_quit)
			sleep(1);
	}
	if (firewall_cfg.mode == FIREWALL_MODE_INTERACTIVE) {
		register_actions_on_flow_parser();
		flow_parser_init("FIREWALL>> ");
		flow_parser_cleanup();
	}

	doca_flow_grpc_destroy();
	/* ARGP cleanup */
	doca_argp_destroy();

	return 0;
}
